﻿Imports System.Device.Location
Imports System.IO.IsolatedStorage
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media
Imports System.Windows.Navigation

Imports Microsoft.Devices.Sensors
Imports Microsoft.Phone.Controls
Imports Microsoft.Phone.Shell
Imports Microsoft.Xna.Framework
Imports Microsoft.Xna.Framework.Input.Touch

Imports OrientingMap.ImageryService

Namespace OrientingMap
	Partial Public Class MainPage
		Inherits PhoneApplicationPage
		Private Const BITRES As Integer = 29

		Private geoWatcher As New GeoCoordinateWatcher(GeoPositionAccuracy.High)
		Private motion As Motion

		Private roadUriTemplate, aerialUriTemplate As String
		Private mapStyle As MapStyle
		Private minimumLevel, maximumLevel As Integer

        Private centerRelativeLongitude As Double
		Private centerRelativeLatitude As Double
		Private locationRelativeLongitude As Double
		Private locationRelativeLatitude As Double

		Private baseLevel As Integer
		Private canvasScale As Double = 1
		Private zoomFactor As Double
		Private currentMapTiles As New Dictionary(Of QuadKey, MapTile)
		Private validQuadKeys As New List(Of QuadKey)
		Private deleteList As New List(Of QuadKey)
		Private canvasOrigin As Vector2
		Private accumulatedDeltaX, accumulatedDeltaY As Double

		Public Sub New()
			InitializeComponent()

			' Fix that annoying ApplicationBar bug
			gotoMeAppBarButton = TryCast(Me.ApplicationBar.Buttons(1), ApplicationBarIconButton)

			' Enable XNA TouchPanel gestures
			TouchPanel.EnabledGestures = GestureType.FreeDrag Or GestureType.DragComplete Or GestureType.Pinch Or GestureType.PinchComplete

			' Initialize the Motion sensor
			If Motion.IsSupported Then
				motion = New Motion
				motion.TimeBetweenUpdates = TimeSpan.FromMilliseconds(33)
				AddHandler motion.CurrentValueChanged, AddressOf OnMotionCurrentValueChanged
			End If

			' Two more event handlers
			AddHandler geoWatcher.PositionChanged, AddressOf OnGeoWatcherPositionChanged
			AddHandler Loaded, AddressOf OnMainPageLoaded
		End Sub

		Protected Overrides Sub OnNavigatedTo(ByVal args As NavigationEventArgs)
			' Get saved information
			Dim settings As IsolatedStorageSettings = IsolatedStorageSettings.ApplicationSettings

			' Initialize to 2010 mean center of US population
			If Not settings.TryGetValue(Of Double)("Longitude", centerRelativeLongitude) Then
				centerRelativeLongitude = GetRelativeLongitude(-90.18471)
			End If

			If Not settings.TryGetValue(Of Double)("Latitude", centerRelativeLatitude) Then
				centerRelativeLatitude = GetRelativeLatitude(38.62452)
			End If

			If Not settings.TryGetValue(Of Double)("ZoomFactor", zoomFactor) Then
				zoomFactor = 12
			End If

			If Not settings.TryGetValue(Of MapStyle)("MapStyle", mapStyle) Then
				mapStyle = MapStyle.Road
			End If

			' Start the Motion sensor and GPS watcher
			If motion IsNot Nothing Then
				Try
					motion.Start()
				Catch
				End Try
			End If

			geoWatcher.Start()
			MyBase.OnNavigatedTo(args)
		End Sub

		Protected Overrides Sub OnNavigatedFrom(ByVal args As NavigationEventArgs)
			' Stop the Motion sensor and GPS watcher
			If motion IsNot Nothing Then
				motion.Stop()
			End If

			geoWatcher.Stop()

			' Save information
			Dim settings As IsolatedStorageSettings = IsolatedStorageSettings.ApplicationSettings
			settings("Longitude") = centerRelativeLongitude
			settings("Latitude") = centerRelativeLatitude
			settings("ZoomFactor") = zoomFactor
			settings("MapStyle") = mapStyle
			settings.Save()

			MyBase.OnNavigatedFrom(args)
		End Sub

		Private Sub OnMainPageLoaded(ByVal sender As Object, ByVal e As RoutedEventArgs)
			' Get the Canvas (0,0) origin for touch processing
			Dim ptOrigin As System.Windows.Point = imageCanvas.TransformToVisual(Nothing).Transform(New System.Windows.Point(0, 0))
			canvasOrigin = New Vector2(CSng(ptOrigin.X), CSng(ptOrigin.Y))

			' Initialize the Bing Maps imagery service
			Dim imageryServiceClient As New ImageryServiceClient("BasicHttpBinding_IImageryService")
			AddHandler imageryServiceClient.GetImageryMetadataCompleted, AddressOf GetImageryMetadataCompleted

			' Make a request for the road metadata
			Dim request As ImageryMetadataRequest = New ImageryMetadataRequest With {.Credentials = New Credentials With {.ApplicationId = "put your own Bing credentials GUID here"}, .Style = MapStyle.Road}
			imageryServiceClient.GetImageryMetadataAsync(request, "road")

			' Make a request for the aerial metadata
			request.Style = MapStyle.Aerial
			imageryServiceClient.GetImageryMetadataAsync(request, "aerial")
		End Sub

		Private Sub GetImageryMetadataCompleted(ByVal sender As Object, ByVal args As GetImageryMetadataCompletedEventArgs)
			If (Not args.Cancelled) AndAlso args.Error Is Nothing Then
				' Get the "powered by" bitmap
				poweredByBitmap.UriSource = args.Result.BrandLogoUri
				poweredByDisplay.Visibility = Visibility.Visible

				' Get the range of map levels available
				Dim result As ImageryMetadataResult = args.Result.Results(0)
				minimumLevel = result.ZoomRange.From
				maximumLevel = result.ZoomRange.To

				' Get the URI and make some substitutions
				Dim uri As String = result.ImageUri
				uri = uri.Replace("{subdomain}", result.ImageUriSubdomains(0))
				uri = uri.Replace("&token={token}", "")
				uri = uri.Replace("{culture}", "en-US")

				If TryCast(args.UserState, String) = "road" Then
					roadUriTemplate = uri
				Else
					aerialUriTemplate = uri
				End If

				' If both URIs are available, let's begin
				If roadUriTemplate IsNot Nothing AndAlso aerialUriTemplate IsNot Nothing Then
					QuadKey.UriTemplate = If(mapStyle = MapStyle.Road, roadUriTemplate, aerialUriTemplate)

					' Calculate baselevel and canvasScale from zoomFactor
                    baseLevel = CInt(zoomFactor)
					canvasScale = Math.Pow(2, zoomFactor - baseLevel)
					imageCanvasScale.ScaleX = canvasScale
					imageCanvasScale.ScaleY = canvasScale

					' Create the initial tiles
					RefreshDisplay()
				End If

				' Use CompositionTarget.Rendering for touch processing
				AddHandler CompositionTarget.Rendering, AddressOf OnCompositionTargetRendering
			Else
				errorTextBlock.Text = "Cannot access Bing Maps: " & args.Error.Message
			End If
		End Sub

		Private Sub OnMotionCurrentValueChanged(ByVal sender As Object, ByVal args As SensorReadingEventArgs(Of MotionReading))
				' Get the rotation matrix & convert to horizontal coordinates
				' Set the azimuth angle to rotation transforms
			Me.Dispatcher.BeginInvoke(Sub()
				Dim matrix As Microsoft.Xna.Framework.Matrix = args.SensorReading.Attitude.RotationMatrix
				Dim vector As Vector3 = Vector3.Transform(Vector3.UnitY, matrix)
				Dim azimuth As Double = 180 * Math.Atan2(vector.X, vector.Y) / Math.PI
				imageCanvasRotate.Angle = azimuth
				northArrowRotate.Angle = azimuth
			End Sub)
		End Sub

		Private Sub OnGeoWatcherPositionChanged(ByVal sender As Object, ByVal args As GeoPositionChangedEventArgs(Of GeoCoordinate))
			' Save the longitude and latitude, set set the location on the display
			locationRelativeLongitude = GetRelativeLongitude(args.Position.Location.Longitude)
			locationRelativeLatitude = GetRelativeLatitude(args.Position.Location.Latitude)
			SetLocationPosition()

			' For the first time this event is fired:
			' Make the location circle visible
			locationDisplay.Visibility = Visibility.Visible

			' And enable the app bar button to move to location
			gotoMeAppBarButton.IsEnabled = True
		End Sub

		Private Sub OnToggleViewAppBarButtonClick(ByVal sender As Object, ByVal args As EventArgs)
			' Toggle the map style
			If mapStyle = MapStyle.Road Then
				mapStyle = MapStyle.Aerial
			Else
				mapStyle = MapStyle.Road
			End If

			' Set the URI template
			QuadKey.UriTemplate = If(mapStyle = MapStyle.Road, roadUriTemplate, aerialUriTemplate)

			' Refresh the tiles with the new map style
			For Each quadKey As QuadKey In currentMapTiles.Keys
				currentMapTiles(quadKey).Refresh()
			Next quadKey
		End Sub

		Private Sub OnGotoMeAppBarButtonClick(ByVal sender As Object, ByVal args As EventArgs)
			centerRelativeLongitude = locationRelativeLongitude
			centerRelativeLatitude = locationRelativeLatitude

			imageCanvas.Children.Clear()
			currentMapTiles.Clear()
			RefreshDisplay()
		End Sub

		' Touch processing using XNA TouchPanel
		Private Sub OnCompositionTargetRendering(ByVal sender As Object, ByVal args As EventArgs)
			Do While TouchPanel.IsGestureAvailable
				Dim gesture As GestureSample = TouchPanel.ReadGesture()

				Select Case gesture.GestureType
					Case GestureType.FreeDrag
						' Adjust delta for rotation of canvas
						Dim delta As Vector2 = TransformGestureToMap(gesture.Delta)

						' Translate the canvas
						imageCanvasTranslate.X += delta.X
						imageCanvasTranslate.Y += delta.Y

						' Adjust the center longitude and latitude
						centerRelativeLongitude -= delta.X / (1 << baseLevel + 8) / canvasScale
						centerRelativeLatitude -= delta.Y / (1 << baseLevel + 8) / canvasScale

						' Accumulate the panning distance
						accumulatedDeltaX += delta.X
						accumulatedDeltaY += delta.Y

						' Check if that's sufficient to warrant a screen refresh
						If Math.Abs(accumulatedDeltaX) > 256 OrElse Math.Abs(accumulatedDeltaY) > 256 Then
							RefreshDisplay()
							accumulatedDeltaX = 0
							accumulatedDeltaY = 0
						End If

					Case GestureType.DragComplete
						Cleanup()

					Case GestureType.Pinch
						' Get the old and new finger positions relative to canvas origin
						Dim newPoint1 As Vector2 = gesture.Position - canvasOrigin
						Dim oldPoint1 As Vector2 = newPoint1 - gesture.Delta

						Dim newPoint2 As Vector2 = gesture.Position2 - canvasOrigin
						Dim oldPoint2 As Vector2 = newPoint2 - gesture.Delta2

						' Rotate in accordance with the current rotation angle
						oldPoint1 = TransformGestureToMap(oldPoint1)
						newPoint1 = TransformGestureToMap(newPoint1)
						oldPoint2 = TransformGestureToMap(oldPoint2)
						newPoint2 = TransformGestureToMap(newPoint2)

						' Find the delta scaling factor
						Dim deltaScale As Double = (newPoint2 - newPoint1).Length() / (oldPoint2 - oldPoint1).Length()

						' But limit it so zoomFactor ranges from 1 to 21
						Dim minDeltaScale As Double = Math.Pow(2, minimumLevel - baseLevel) / canvasScale
						Dim maxDeltaScale As Double = Math.Pow(2, maximumLevel - baseLevel) / canvasScale
						deltaScale = Math.Min(maxDeltaScale, Math.Max(minDeltaScale, deltaScale))

						' Find oldScale and newScale
						Dim oldScale As Double = canvasScale
						Dim newScale As Double = oldScale * deltaScale

						' Use that to adjust the Canvas scale
						canvasScale *= deltaScale
						imageCanvasScale.ScaleX = canvasScale
						imageCanvasScale.ScaleY = canvasScale

						' Calculate the new zoomFactor
						zoomFactor = baseLevel + Math.Log(canvasScale) / Math.Log(2)

						' Now find the translation factor implied by the gesture deltas
						Dim oldTranslate As New Vector2(CSng(imageCanvasTranslate.X), CSng(imageCanvasTranslate.Y))
						Dim translate1 As Vector2 = newPoint1 - CSng(deltaScale) * (oldPoint1 - oldTranslate)
						Dim translate2 As Vector2 = newPoint2 - CSng(deltaScale) * (oldPoint2 - oldTranslate)

						' Take the average and use that for the new translation factors
						Dim newTranslate As Vector2 = 0.5f * (translate1 + translate2)
						imageCanvasTranslate.X = newTranslate.X
						imageCanvasTranslate.Y = newTranslate.Y

						' Adjust the center longitude and latitude similarly
						centerRelativeLongitude -= newTranslate.X / (1 << baseLevel + 8) / newScale - oldTranslate.X / (1 << baseLevel + 8) / oldScale
						centerRelativeLatitude -= newTranslate.Y / (1 << baseLevel + 8) / newScale - oldTranslate.Y / (1 << baseLevel + 8) / oldScale

						' If the Canvas has doubled in size, shift the scaling to the tiles
						Do While canvasScale >= 2
							baseLevel += 1
							canvasScale /= 2
							imageCanvasScale.ScaleX = canvasScale
							imageCanvasScale.ScaleY = canvasScale

							For Each child As UIElement In imageCanvas.Children
								TryCast(child, MapTile).Scale *= 2
								Canvas.SetLeft(child, 2 * Canvas.GetLeft(child))
								Canvas.SetTop(child, 2 * Canvas.GetTop(child))
							Next child
						Loop

						' Also if the Canvas has halved its size
						Do While canvasScale <= 0.5
							baseLevel -= 1
							canvasScale *= 2
							imageCanvasScale.ScaleX = canvasScale
							imageCanvasScale.ScaleY = canvasScale

							For Each child As UIElement In imageCanvas.Children
								TryCast(child, MapTile).Scale /= 2
								Canvas.SetLeft(child, Canvas.GetLeft(child) / 2)
								Canvas.SetTop(child, Canvas.GetTop(child) / 2)
							Next child
						Loop

						RefreshDisplay()

					Case GestureType.PinchComplete
						Cleanup()
				End Select
			Loop
		End Sub

		Private Function TransformGestureToMap(ByVal vector As Vector2) As Vector2
			Dim radians As Double = Math.PI * imageCanvasRotate.Angle / 180
			Dim x As Double = vector.X * Math.Cos(radians) + vector.Y * Math.Sin(radians)
			Dim y As Double = -vector.X * Math.Sin(radians) + vector.Y * Math.Cos(radians)
			Return New Vector2(CSng(x), CSng(y))
		End Function

		Private Sub RefreshDisplay()
			validQuadKeys.Clear()

			' Find integer longitude and latitude
            Dim centerLongitude As Integer = CInt(centerRelativeLongitude * (1 << BITRES))
            Dim centerLatitude As Integer = CInt(centerRelativeLatitude * (1 << BITRES))

			' Crop them based on the base level, and get QuadKey
			Dim longitude As Integer = centerLongitude >> BITRES - baseLevel
			Dim latitude As Integer = centerLatitude >> BITRES - baseLevel
			Dim baseQuadKey As New QuadKey(longitude, latitude, baseLevel)

			' Calculate offsets based on location within tile
			Dim xOffset As Double = (centerLongitude >> BITRES - baseLevel - 8) Mod 256
			xOffset = (centerLongitude / CDbl(1 << BITRES - baseLevel - 8)) Mod 256

			Dim yOffset As Double = (centerLatitude >> BITRES - baseLevel - 8) Mod 256
			yOffset = (centerLatitude / CDbl(1 << BITRES - baseLevel - 8)) Mod 256

			xOffset += imageCanvasTranslate.X / canvasScale
			yOffset += imageCanvasTranslate.Y / canvasScale

			' Number of tiles to create depends on current canvasScale
            Dim scope As Integer = CInt(Math.Ceiling(3 / canvasScale))

			For row As Integer = -scope To scope
				For col As Integer = -scope To scope
					Dim quadKey As QuadKey = baseQuadKey.GetQuadKeyFromRowColumnOffsets(col, row)

					If quadKey Is Nothing Then
						Continue For
					End If

					validQuadKeys.Add(quadKey)

					If Not currentMapTiles.ContainsKey(quadKey) Then
						Dim x As Double = 256 * col - xOffset
						Dim y As Double = 256 * row - yOffset

						x = Math.Round(x)
						y = Math.Round(y)

						' Create MapTile
						Dim mapTile As MapTile = New MapTile With {.ImageOpened = AddressOf OnMapTileImageOpened, .QuadKey = quadKey}

						Canvas.SetLeft(mapTile, x)
						Canvas.SetTop(mapTile, y)
						Canvas.SetZIndex(mapTile, quadKey.Level)
						imageCanvas.Children.Add(mapTile)
						currentMapTiles.Add(quadKey, mapTile)
					End If
				Next col
			Next row

			SetLocationPosition()
		End Sub

		Private Sub Cleanup()
			deleteList.Clear()

			For Each quadKey As QuadKey In currentMapTiles.Keys
				If quadKey.Level = baseLevel Then
					If Not validQuadKeys.Contains(quadKey) Then
						deleteList.Add(quadKey)
					End If
				ElseIf quadKey.Level < baseLevel Then
					Dim tileStillVisible As Boolean = False

					For Each childQuadKey As QuadKey In quadKey.Children
						If currentMapTiles.ContainsKey(childQuadKey) AndAlso (Not currentMapTiles(childQuadKey).IsImageOpened) Then
							tileStillVisible = True
							Exit For
						End If
					Next childQuadKey

					If Not tileStillVisible Then
						deleteList.Add(quadKey)
					End If
				Else
					Dim parentQuadKey As QuadKey = quadKey.Parent

					If (Not currentMapTiles.ContainsKey(parentQuadKey)) OrElse currentMapTiles(parentQuadKey).IsImageOpened Then
						deleteList.Add(quadKey)
					End If
				End If
			Next quadKey

			For Each quadKey As QuadKey In deleteList
				Dim mapTile As MapTile = currentMapTiles(quadKey)
				mapTile.QuadKey = Nothing
				mapTile.ImageOpened = Nothing

				imageCanvas.Children.Remove(mapTile)
				currentMapTiles.Remove(quadKey)
			Next quadKey
		End Sub

		Private Sub OnMapTileImageOpened(ByVal sender As Object, ByVal args As EventArgs)
			Cleanup()
		End Sub

		' Also called from geo watcher
		Private Sub SetLocationPosition()
			locationTranslate.X = 256 * (1 << baseLevel) * (locationRelativeLongitude - centerRelativeLongitude) * imageCanvasScale.ScaleX - imageCanvasTranslate.X
			locationTranslate.Y = 256 * (1 << baseLevel) * (locationRelativeLatitude - centerRelativeLatitude) * imageCanvasScale.ScaleY - imageCanvasTranslate.Y
		End Sub

		Private Function GetRelativeLongitude(ByVal longitude As Double) As Double
			Return (180 + longitude) / 360
		End Function

		Private Function GetRelativeLatitude(ByVal latitude As Double) As Double
			Dim sin As Double = Math.Sin(Math.PI * latitude / 180)
			Return 0.5 - Math.Log((1 + sin) / (1 - sin)) / (4 * Math.PI)
		End Function
	End Class
End Namespace